
/*
	function interface for communication with NIC driver
*/

#include <linux/config.h>
#include <linux/version.h>
#include "kcompat.h"

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/rtnetlink.h>

#include "vmns.h"
#include "drv_func.h"
#include "vmns_drv.h"
#include "vmns_mem.h"
#include "debug.h"

extern BOOL	bInIoctl;

extern spinlock_t	vmns_lock, vmns_event_lock;

static void free_mc_list(struct dev_mc_list**,int *);

static int vmns_dev_open(PVMNS_DEVICE vdev);
static int vmns_dev_close(PVMNS_DEVICE vdev);
static int vmns_dev_ioctl(PVMNS_DEVICE vdev,struct ifreq*,int);
static int vmns_dev_send(PVMNS_DEVICE vdev, PVMNS_SKB vskb);
static int vmns_dev_set_mc_list(PVMNS_DEVICE vdev, PVMNS_MC_LIST mc_list , int rx_flags);
static PVMNS_DEVICE_STATS vmns_dev_get_stats(PVMNS_DEVICE vdev);

#ifdef LINUX_KERNEL_24X
BOOL	vmns_is_24X=TRUE;
#else
BOOL	vmns_is_24X=FALSE;
#endif

#ifdef CONFIG_PROC_FS
BOOL	vmns_have_procfs=TRUE;
#else
BOOL	vmns_have_procfs=FALSE;
#endif

PVMNS_DEVICE init_vmns_device(char* name) {

	
	PVMNS_DEVICE	pvdev=NULL;		
	struct net_device* dev;
	
	if ((dev=dev_get_by_name(name))==NULL) {
		return pvdev;
	}
//	dev->flags|=IFF_NOARP;	
	pvdev=VMNS_MALLOC(sizeof(VMNS_DEVICE));
		
	if (pvdev==NULL) 
		return pvdev;
	memset(pvdev,0,sizeof(VMNS_DEVICE));

	pvdev->pOsDev=dev;
	pvdev->name=(char*)dev->name;
	memcpy(pvdev->mac.addr,dev->dev_addr,6);
	
	pvdev->open=vmns_dev_open;
	pvdev->close=vmns_dev_close;
	
	pvdev->send=vmns_dev_send;
	pvdev->get_stats=vmns_dev_get_stats;
	pvdev->set_mc_list=vmns_dev_set_mc_list;
	pvdev->ioctl=vmns_dev_ioctl;
	pvdev->stats=VMNS_MALLOC(sizeof(VMNS_DEVICE_STATS));
	memset(pvdev->stats,0,sizeof(VMNS_DEVICE_STATS));
#ifdef LINUX_KERNEL_24X	
	dev_put(dev);	
#endif	
	return pvdev;
}

static int vmns_dev_open(PVMNS_DEVICE vdev)
{
	struct net_device* dev=VMNS_GET_DEVICE(vdev);
	int res;
		
	if (!bInIoctl) 
		rtnl_lock();
		
	res=dev_open (dev);
	
	if (!bInIoctl) 
		rtnl_unlock();
		
	return res;
}

static int vmns_dev_close(PVMNS_DEVICE vdev)
{
	struct net_device* dev=VMNS_GET_DEVICE(vdev);
	int res;
	
	if (!bInIoctl) 
		rtnl_lock();
		
	res=dev_close (dev);
	
	if (!bInIoctl) 
		rtnl_unlock();
		
	return res;
}

static int vmns_dev_send(PVMNS_DEVICE vdev, PVMNS_SKB vskb) {
	struct net_device* dev=VMNS_GET_DEVICE(vdev);
	struct sk_buff* skb=VMNS_GET_SKB(vskb);
	
	ASSERT(vdev && vskb);	
			
	skb->dev=dev;
		
	dev_queue_xmit(skb);
	
	return VMNS_OK;
}

int vmns_dev_set_mac(PVMNS_DEVICE vdev, void* addr) {	
	return -EBUSY;
}

int vmns_netif_rx(PVMNS_DEVICE vdev, PVMNS_SKB vskb) {
	struct sk_buff* skb=VMNS_GET_SKB(vskb);
	struct net_device* dev=VMNS_GET_DEVICE(vdev);
	
	ASSERT(vdev && vskb);
	skb->dev=dev;
	netif_rx(skb);
	return VMNS_OK;	
}

static 
PVMNS_DEVICE_STATS vmns_dev_get_stats(PVMNS_DEVICE vdev) {
	struct net_device* dev=VMNS_GET_DEVICE(vdev);
	PVMNS_DEVICE_STATS			pvstats=VMNS_GET_STATS(vdev);
	
	
	if (dev->get_stats) {
		struct net_device_stats*	pstats;
		
		pstats=dev->get_stats(dev);
		
		ASSERT(pstats && pvstats);	
		
		pvstats->rx_packets=pstats->rx_packets;
		pvstats->tx_packets=pstats->tx_packets;
		pvstats->rx_bytes=pstats->rx_bytes;	
		pvstats->tx_bytes=pstats->tx_bytes;
		pvstats->multicast=pstats->multicast;
		pvstats->collisions=pstats->collisions;
	}
	return pvstats;
}

static int 
vmns_dev_set_mc_list(PVMNS_DEVICE vdev, PVMNS_MC_LIST mc_list , int rx_flags) {
	struct net_device* 	dev=VMNS_GET_DEVICE(vdev);
	struct dev_mc_list* tmp;
	PVMNS_MC_ENTRY		pmc_entry;

	
	if (dev->set_multicast_list==NULL) {
		return VMNS_FAIL;
	}
	

	free_mc_list(&dev->mc_list,&dev->mc_count);

	for (pmc_entry=mc_list->mc_list;pmc_entry;pmc_entry=pmc_entry->next) {
		tmp=kmalloc(sizeof(struct dev_mc_list),GFP_ATOMIC);
		memcpy(tmp->dmi_addr,pmc_entry->mac.addr,6);	
		tmp->dmi_addrlen=pmc_entry->dmi_addrlen;
		tmp->dmi_users=pmc_entry->dmi_users;	
		tmp->dmi_gusers=pmc_entry->dmi_gusers;
		tmp->next=dev->mc_list;
		dev->mc_list=tmp;
		dev->mc_count++;
	}

	if (rx_flags & VMNS_DEV_PROMISC)
		dev->flags |= IFF_PROMISC;
	else
		dev->flags &= ~IFF_PROMISC;
		
	if (rx_flags & VMNS_DEV_ALLMULTI)	
		dev->flags |= IFF_ALLMULTI;
	else
		dev->flags &= ~IFF_ALLMULTI;
	
	dev->set_multicast_list(dev);
	
	free_mc_list(&dev->mc_list,&dev->mc_count);	
	
	return VMNS_OK;	
}

static void free_mc_list(struct dev_mc_list** ppmc_list,int *count) {
	struct dev_mc_list* ptr=*ppmc_list;
	while (*ppmc_list) {
		ptr=*ppmc_list;
		*ppmc_list=(*ppmc_list)->next;
		kfree(ptr);
	}	
	*ppmc_list=NULL;*count=0;
}

void free_vmns_mc_list(PVMNS_MC_LIST pvmc_list) {
	PVMNS_MC_ENTRY pvmc_entry=pvmc_list->mc_list;
	PVMNS_MC_ENTRY	ptr;
	while (pvmc_entry) {
		ptr=pvmc_entry;
		pvmc_entry=pvmc_entry->next;
		VMNS_FREE(ptr,sizeof(VMNS_MC_ENTRY));	
	}	
	pvmc_list->mc_list=NULL;pvmc_list->num=0;
}



int vmns_dev_ioctl(PVMNS_DEVICE vdev,struct ifreq* ifreq, int pt) {
	struct net_device* dev=vdev->pOsDev;
	if (dev->do_ioctl) {
		if (dev->do_ioctl(dev,ifreq,pt)!=0) 
			return VMNS_FAIL;
	}
	return VMNS_OK;
}

void put_vmns_device(PVMNS_DEVICE* pvdev) {
	ASSERT((*pvdev) && (*pvdev)->stats);
#ifdef LINUX_KERNEL_24X	
	dev_put((*pvdev)->pOsDev);
#endif	
	VMNS_FREE(*pvdev,sizeof(VMNS_DEVICE));	
	VMNS_FREE((*pvdev)->stats,sizeof(VMNS_DEVICE_STATS));
	*pvdev=NULL;
}

void free_vmns_device(PVMNS_DEVICE* pvdev) {
//	struct net_device* dev=VMNS_GET_DEVICE(*pvdev);
	ASSERT((*pvdev) && (*pvdev)->stats);
//	dev->flags&=~IFF_NOARP;
	
	if (*pvdev)	{
		if ((*pvdev)->stats) 
			VMNS_FREE((*pvdev)->stats,sizeof(VMNS_DEVICE_STATS));	
			
		VMNS_FREE(*pvdev,sizeof(VMNS_DEVICE));
	}
	*pvdev=NULL;	
}

int vmns_strncmp(char* str1,char* str2,int size) {
	ASSERT(str1&&str2);
	return strncmp(str1,str2,size);
}

char* vmns_strncpy(char* str1,char* str2,int size) {
	ASSERT(str1&&str2);	
	return strncpy(str1,str2,size);
}

void* vmns_memcpy(void* p1,void* p2, int size) {
	ASSERT(p1&& p2);
	return memcpy(p1,p2,size);
}

void* vmns_memset(void* p1,int p, int size) {
	ASSERT(p1);
	return memset(p1,p,size);
}
int vmns_memcmp(void* p1,void* p2, int size) {
	ASSERT(p1 && p2);
	return memcmp(p1,p2,size);
}

int	vmns_strlen(char* p) {
	ASSERT(p);
	return strlen(p);
}

ULONG vmns_get_hz(void) {
	return (ULONG) HZ;
}

void vmns_registerdev(PVMNS_DEVICE pvdev) {
	struct net_device* dev=VMNS_GET_DEVICE(pvdev);
	ASSERT(pvdev);	
	register_netdevice(dev);			
}

void vmns_send_packet(PVMNS_DEVICE pvdev,unsigned char* dst, unsigned char* src, 
	unsigned int type,unsigned char* pPacket,int size) {
	
	struct sk_buff *skb;
	unsigned char*	ptr;
	struct net_device* dev=pvdev->pOsDev;
	
	skb=dev_alloc_skb(size+16);
	
	if (skb==NULL)
		return;
		
	skb_reserve(skb, (dev->hard_header_len+15)&~15);
	skb->protocol=type;
	skb->nh.raw=skb->data;
	skb->dev=pvdev->pOsDev;
	
	ptr = skb_put(skb,size);
	
	if (pPacket!=NULL)	
		memcpy(ptr,pPacket,size);
		
	if (dst==NULL)
		dst=dev->broadcast;
	if (src==NULL)
		src=dev->dev_addr;		 		
		
	if (dev->hard_header &&
		dev->hard_header(skb,dev,type,dst,src,skb->len) < 0) {
		dev_kfree_skb(skb);
		return;
	}
	dev_queue_xmit(skb);
}

void vmns_kern_lock(void) {
	kern_sp_lock(&vmns_lock);
}

void vmns_kern_unlock(void) {
	kern_sp_unlock(&vmns_lock);
}

void vmns_kern_lock_bh(void) {
	kern_sp_lock_bh(&vmns_lock);
}

void vmns_kern_unlock_bh(void) {
	kern_sp_unlock_bh(&vmns_lock);
}

void vmns_kern_event_lock(BOOL irq,int *pflags) {
	if (irq) 
		kern_sp_lock_irq(&vmns_event_lock,(*pflags));
	else
		kern_sp_lock(&vmns_event_lock);
}

void vmns_kern_event_unlock(BOOL irq,int *pflags) {
	if (irq) 
		kern_sp_unlock_irq(&vmns_event_lock,(*pflags));
	else
		kern_sp_unlock(&vmns_event_lock);
}

void vmns_if_stop(PVMNS_DEVICE pvdev) {
	struct net_device* dev=VMNS_GET_DEVICE(pvdev);
	netif_stop_queue(dev);	
}

int vmns_sprintf(char* str,const char *fmt, ...) {
	int i;
	va_list args;
	va_start(args, fmt);
	i=vsprintf(str, fmt, args);
	va_end(args);
	return i;
}
void vmns_if_wake(PVMNS_DEVICE pvdev) {
	struct net_device* dev=VMNS_GET_DEVICE(pvdev);
	netif_wake_queue(dev);	
}

