/*
	Author	:	Shing
	Date	:	2001/04/24
*/

#include <linux/version.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/ioctl.h>
#include <linux/socket.h>
#include <linux/rtnetlink.h>
#include <linux/if.h>
#include <asm/unaligned.h>

#include "kcompat.h"
#include "vmns_cfg.h"
#include "vmns.h"
#include "debug.h"
#include "vmns_vadp.h"
#include "vmns_mem.h"
#include "vmns_drv.h"
#include "drv_func.h"

static int vadp_send(struct sk_buff *skb, struct net_device *dev);
static int vadp_open(struct net_device *vdev);
static int vadp_close(struct net_device *vdev);
static int vadp_ioctl(struct net_device *vdev, struct ifreq *ifr, int cmd);
static int vadp_set_mac(struct net_device*dev, void* paddr);
static void vadp_set_mc_list(struct net_device *dev);
static struct net_device_stats* vadp_get_stats(struct net_device *dev);

#define IS_UP(dev)	((((dev)->flags & (IFF_UP)) == (IFF_UP)) && \
                     (netif_running(dev) && netif_carrier_ok(dev)))
extern BOOL bInIoctl;

#ifdef LINUX_KERNEL_24X
extern spinlock_t vmns_lock;
#endif
                    
PVMNS_DEVICE init_vadp_device(char* name) {
	struct net_device *new_dev;
	PVMNS_DEVICE	pvdev;
		
	new_dev=vmnsInitEtherdev(name);
    if (new_dev != NULL) {
        new_dev->open 				= 	vadp_open;
        new_dev->stop 				= 	vadp_close;
        new_dev->do_ioctl			=	vadp_ioctl;
	    new_dev->get_stats			=	vadp_get_stats;
		new_dev->priv				=	VMNS_MALLOC(sizeof(VADP_PRIVATE));
		new_dev->hard_start_xmit	=	vadp_send;
		new_dev->tx_queue_len		=	0;
		new_dev->set_mac_address 	=	vadp_set_mac;
		new_dev->set_multicast_list =	vadp_set_mc_list;
		new_dev->flags				&=	~IFF_UP;
#ifdef LINUX_KERNEL_22X
		new_dev->start				=	0;
#endif		
		if (new_dev->priv) {		
			memset(new_dev->priv,0,sizeof(VADP_PRIVATE));
		}
   } else {   
   		return NULL;
   }
   
   pvdev=VMNS_MALLOC(sizeof(VMNS_DEVICE));
   memset(pvdev,0,sizeof(VMNS_DEVICE));
   if (pvdev) {
   		PVADP_PRIVATE vpriv=VADP_GET_PRIVATE(new_dev);
   		
   		pvdev->pOsDev=new_dev;
   		memcpy(pvdev->mac.addr,new_dev->dev_addr,6);
   		pvdev->dev_addr=new_dev->dev_addr;
   		pvdev->stats=VMNS_MALLOC(sizeof(VMNS_DEVICE_STATS));
   		memset(pvdev->stats,0,sizeof(VMNS_DEVICE_STATS));
        vpriv->pDev=pvdev; 
        vpriv->stats=VMNS_MALLOC(sizeof(struct net_device_stats));       
        memset(vpriv->stats,0,sizeof(struct net_device_stats));
   }
   return pvdev;
}


int unregister_vadp_device(PVMNS_DEVICE pvdev) {
	struct net_device* dev=pvdev->pOsDev;
	ASSERT(dev && pvdev && pvdev->pOsDev);
	
	if (dev!=NULL) {
		PVADP_PRIVATE vpriv=VADP_GET_PRIVATE(dev);
		
		
		(bInIoctl) ? unregister_netdevice(dev) : unregister_netdev(dev);	

		if (vpriv->pDev) {
			free_vmns_device(&(vpriv->pDev));
		}
		
		if (vpriv->stats)
			VMNS_FREE(vpriv->stats,sizeof(struct net_device_stats));		
		if (vpriv) 
			VMNS_FREE(vpriv,sizeof(VADP_PRIVATE));		
	
		VMNS_FREE(dev,sizeof(struct net_device));
	}
	return 0;	
}

static 
int vadp_open(struct net_device *dev) {
	PVMNS_DEVICE	pvdev=VADP_GET_DEVICE(dev);	
	ASSERT(pvdev && dev);

	kern_sp_lock_bh(&vmns_lock);

	if (pvdev->open) {
		if (pvdev->open(pvdev)==VMNS_OK) {
			MOD_INC_USE_COUNT;
#ifdef LINUX_KERNEL_22X
            dev->start = 1;
#endif       
			kern_sp_unlock_bh(&vmns_lock);
			return 0;     			
		}
		kern_sp_unlock_bh(&vmns_lock);		
		return -EPERM;
	}
	kern_sp_unlock_bh(&vmns_lock);
	return -EPERM;
}

static
int vadp_close(struct net_device *dev) {
	PVMNS_DEVICE	pvdev=VADP_GET_DEVICE(dev);	

	ASSERT(pvdev && dev);

	kern_sp_lock_bh(&vmns_lock);

	if (pvdev->close) {
		if (pvdev->close(pvdev)==VMNS_OK) {
			MOD_DEC_USE_COUNT;
#ifdef LINUX_KERNEL_22X
            dev->start = 0;
#endif            	
			kern_sp_unlock_bh(&vmns_lock);
			return 0;		
		}
		kern_sp_unlock_bh(&vmns_lock);
		return -EPERM;
	}

	kern_sp_unlock_bh(&vmns_lock);
	return -EPERM;
}

static
int vadp_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd) {
	PVMNS_DEVICE	pvdev=VADP_GET_DEVICE(dev);

	ASSERT(pvdev);
	kern_sp_lock_bh(&vmns_lock);	
	
	if ((cmd >= SIOCDEVPRIVATE) && (cmd < (SIOCDEVPRIVATE+16))) {
		kern_sp_unlock_bh(&vmns_lock);
        return 0;
    }	

	if (pvdev && pvdev->ioctl) {
		if (pvdev->ioctl(pvdev,ifr,cmd)==VMNS_OK) {
			kern_sp_unlock_bh(&vmns_lock);
			return 0;
		}		
		kern_sp_unlock_bh(&vmns_lock);
		return -EBUSY;					
	}	
	kern_sp_unlock_bh(&vmns_lock);
	return -EBUSY;
}

static void vadp_set_mc_list(struct net_device *dev) {
	PVMNS_DEVICE		pvdev=VADP_GET_DEVICE(dev);	
	VMNS_MC_LIST		vmc_list={0,NULL};
	struct	dev_mc_list*	pdmc_list;

	int				flags=0;


	kern_sp_lock_bh(&vmns_lock);
	
	if (pvdev->set_mc_list)	{
	
		for (pdmc_list=dev->mc_list;pdmc_list;pdmc_list=pdmc_list->next) {
			PVMNS_MC_ENTRY	ptr=VMNS_MALLOC(sizeof(VMNS_MC_ENTRY));
			memcpy(ptr->mac.addr,pdmc_list->dmi_addr,6);
			ptr->dmi_addrlen=pdmc_list->dmi_addrlen;
			ptr->dmi_users=pdmc_list->dmi_users;
			ptr->dmi_gusers=pdmc_list->dmi_gusers;	
			ptr->next=vmc_list.mc_list;
			vmc_list.mc_list=ptr;
			vmc_list.num++;	
		}							
										
		if (dev->flags & IFF_PROMISC)
			flags |= VMNS_DEV_PROMISC;
		else
			flags &= ~VMNS_DEV_PROMISC;
		
		if (dev->flags & IFF_ALLMULTI)	
			flags |= VMNS_DEV_ALLMULTI;
		else
			flags &= ~VMNS_DEV_ALLMULTI;
	
		pvdev->set_mc_list(pvdev,&vmc_list,flags);
		
	}
	
	free_vmns_mc_list(&vmc_list);
	kern_sp_unlock_bh(&vmns_lock);
}

static
int vadp_set_mac(struct net_device *dev, void *addr) {
	PVMNS_DEVICE		pvdev=VADP_GET_DEVICE(dev);	
	ASSERT(dev && addr);
	
	kern_sp_lock_bh(&vmns_lock);
	
	if (pvdev->set_mac) {
		if (pvdev->set_mac(pvdev,addr)==VMNS_OK) {
			kern_sp_unlock_bh(&vmns_lock);
			return 0;
		}
	}	
	kern_sp_unlock_bh(&vmns_lock);	
	return -1;
}

static struct net_device_stats*
vadp_get_stats(struct net_device *dev) {
	struct net_device_stats* pstats=VADP_GET_STATS(dev);
	PVMNS_DEVICE		pvdev=VADP_GET_DEVICE(dev);	
	PVMNS_DEVICE_STATS	pvstats=VMNS_GET_STATS(pvdev);
	
	ASSERT(pstats && pvdev && pvstats);
	kern_sp_lock_bh(&vmns_lock);
	
	if (pvdev->get_stats) {
		pvstats=pvdev->get_stats(pvdev);
	}
	pstats->rx_packets=pvstats->rx_packets;
	pstats->tx_packets=pvstats->tx_packets;
	pstats->rx_bytes=pvstats->rx_bytes;	
	pstats->tx_bytes=pvstats->tx_bytes;
	pstats->multicast=pvstats->multicast;
	pstats->collisions=pvstats->collisions;
	
	kern_sp_unlock_bh(&vmns_lock);	
   	return pstats;
   	
}

int vadp_send(struct sk_buff *skb, struct net_device *dev) {
	PVMNS_DEVICE	pvdev=VADP_GET_DEVICE(dev);	
	static VMNS_SKB	vskb;

	ASSERT(pvdev);

	kern_sp_lock_bh(&vmns_lock);

	if (pvdev->send) {
		vskb.pOsSkb=skb;
		vskb.psrc=(PMAC_ADDR)skb->mac.ethernet->h_dest;
		vskb.pdst=(PMAC_ADDR)skb->mac.ethernet->h_source;
		vskb.ptype=&(skb->mac.ethernet->h_proto);
		vskb.pprotocol=&(skb->protocol);
		vskb.phead=skb->head;
		vskb.pdata=skb->data;
		vskb.len=skb->len;
		vskb.pInfo=GET_VMNS_ATTACH_INFO(skb);
		
		if (pvdev->send(pvdev,&vskb)!=VMNS_OK) {
			kern_sp_unlock_bh(&vmns_lock);
			return 1;
		}
	}
	kern_sp_unlock_bh(&vmns_lock);
	return 0;
}
